/*       ͻ
             TITLE:  HDLC State machine routines for Keil C        
          FILENAME:  HDLC.c                                        
          REVISION:  VER 0.0                                       
          REV.DATE:  13-12-04                                      
           ARCHIVE:                                                
            AUTHOR:  Copyright (c) 2004, Murray R.Van Luyn.        
         ͼ        */


#include "HDLC.h"
#include "Crc.h"
#include "Sio_poll.h"  // Testing
#include <stdio.h>     // printf() Testing

/*  ͻ
                           Privately Defined Constants                   
    ͼ  */


/*  ͻ
                             Private Global Variables                    
    ͼ  */
unsigned char hdlc_state = RX_IDLE;             // State machine internal state

// Receiver Global Variables
unsigned char xdata rx_buffer[RX_BUFFER_LEN];   // Circular packet buffer
// Use unsigned int for next 3 indexes when RX_BUFFER_LEN > 255
unsigned char rx_buffer_read_index  = 0;        // Read index
unsigned char rx_buffer_write_index = 0;        // Committed write index
unsigned char rx_buffer_trial_index = 1;        // Uncommitted write index

unsigned char rx_shift_reg = 0;                 // Decoder stream shift reg
unsigned char rx_bit_count = 0;                 // Shift reg bit counter

// Transmitter Global Variables
unsigned char xdata tx_buffer[TX_BUFFER_LEN];   // Circular packet buffer
// Use unsigned int for next 3 indexes when TX_BUFFER_LEN > 255
unsigned char tx_buffer_read_index  = 0;        // Read index
unsigned char tx_buffer_write_index = 0;        // Committed write index
unsigned char tx_buffer_trial_index = 1;        // Uncommitted write index

unsigned int  tx_shift_reg = 0;                 // Encoder stream shift reg
unsigned char tx_bit_count = 0;                 // Shift reg bit counter

unsigned char tx_preamble_cnt = TX_PREAMBLE_LEN;

// Precomputed PCA counter reload values - avoid slow MUL and DIV
code unsigned char compare_reload_L[9] = 
                                  {PCA_1_BIT_COUNT * 0 % 256,
                                   PCA_1_BIT_COUNT * 1 % 256,
                                   PCA_1_BIT_COUNT * 2 % 256,
                                   PCA_1_BIT_COUNT * 3 % 256,
                                   PCA_1_BIT_COUNT * 4 % 256,
                                   PCA_1_BIT_COUNT * 5 % 256,
                                   PCA_1_BIT_COUNT * 6 % 256,
                                   PCA_1_BIT_COUNT * 7 % 256,
                                   PCA_1_BIT_COUNT * 8 % 256};

code unsigned char compare_reload_H[9] = 
                                  {PCA_1_BIT_COUNT * 0 / 256,
                                   PCA_1_BIT_COUNT * 1 / 256,
                                   PCA_1_BIT_COUNT * 2 / 256,
                                   PCA_1_BIT_COUNT * 3 / 256,
                                   PCA_1_BIT_COUNT * 4 / 256,
                                   PCA_1_BIT_COUNT * 5 / 256,
                                   PCA_1_BIT_COUNT * 6 / 256,
                                   PCA_1_BIT_COUNT * 7 / 256,
                                   PCA_1_BIT_COUNT * 8 / 256};


/*  ͻ
                           Private Function Prototypes                   
    ͼ  */
//void hdlc_transition_isr(void) interrupt 6 using 2;
void rx_drop_packet(void);
void rx_process_packets(void);
void rx_add_bit_to_buffer(bit bit_to_add);
void rx_reset_shift_reg(void);
bit  tx_buffer_overrun(void);


/*  ͻ
                             Function Implementation                     
    ͼ  */

/*  ͻ
                                                                         
                                 hdlc_init_pca()                         
                                                                         
                                                                         
                                                                         
    ͼ  */
void hdlc_init_pca(void)
{
    // Start with tx data low and rx data as an input
    TX_PORT_PIN = 0;
    RX_PORT_PIN = 1;

    // Set PCA counter to mode 1. Increment at XTAL / 12 in X1 mode
    // 256kHz with 18.432MHz Xtal in X1 mode
    CMOD = 0x00;

    // Reset PCA counter registers
    CH = 0;
    CL = 0;

    // Load PCA module 1 compare registers with count for compare match
    // when an 8 bit long, no transition period occurs.
    CCAP1L = compare_reload_L[8];
    CCAP1H = compare_reload_H[8];

    // Load PCA module 2 compare registers with count for compare match
    // when an 8 bit long, no transition period occurs.
    CCAP2L = compare_reload_L[8];
    CCAP2H = compare_reload_H[8];

    // Set PCA module 0 to capture all transitions on CEX0 P1.3, and then
    // generate PCA interrupt 6 with vector 33h.
    // ECCF0 interrupt enable bit is set (enabled)
    CCAPM0 = 0x31;

    // Set PCA module 1 to compare PCA counter with compare registers, and 
    // to generate PCA interrupt 6 with vector 33h if the same.
    // ECCF1 interrupt enable bit is set (enabled)
    CCAPM1 = 0x49;

    // Set PCA module 2 to compare PCA counter with compare registers, and 
    // not yet to generate PCA interrupt 6 with vector 33h if the same.
    // ECCF2 interrupt enable bit is reset (inhibited)
    CCAPM2 = 0x48;

    // Set PCA counter running
    CR = 1;

    // Enable PCA interrupt
    EC = 1;

    // Enable all interrupts
    EA = 1;

	// Initialise global memory pool to start at X:0x0300 and occupy 0xFF bytes
    init_mempool ((void xdata *)0x0300, 0x0100);

    // Zero first incoming packet length
    rx_buffer[rx_buffer_write_index] = 0;

    return;
}


/*  ͻ
                                                                         
                              hdlc_transition_isr()                      
                                                                         
                                                                         
                                                                         
    ͼ  */
void hdlc_transition_isr(void) interrupt 6 using 2
{
    unsigned char inter_trans_length = 0;

    // Disable PCA interrupt
    EC = 0;

    // Stop PCA counter running
    CR = 0;

    // Reset PCA counter registers
    CH = 0;
    CL = 0;

    // Restart PCA counter running
    CR = 1;

    // Calculate the length between this transition and the last in bits.
    // Transitions occur when each zero bit is sent.
    // Valid range: 1..7 (0 to 6 ones between zeroes)
    // Invalid range: 8+ (7+ ones between zeroes, or no transition timeout)

    // Check if interrupt was for PCA transition, overflow or match
    if (CCF0)
    {
        // Interrupt due to PCA Module 0 transition

        int mod_0_capture;

        // Reset interrupt source flag
        CCF0 = 0;

        // Calculate transition length from PCA module 0 capture registers.
        // Each bit period takes PCA_1_BIT_COUNT counts. Sampling occurs in 
        // the middle of bit periods, between bits. Very costly routine in
        // terms of time even if we avoid DIV instruction.
        // Routine returns 0 for transitions < 1/2 bit period.
        // This routine takes 97us for 7 bits at 18.432Mhz. (faster for less)
        mod_0_capture = (CCAP0H << 8) + CCAP0L - PCA_1_BIT_COUNT / 2;                                              
        while (mod_0_capture > 0)
        {
            inter_trans_length++;
            mod_0_capture -= PCA_1_BIT_COUNT;
        }

    }
    else if(CCF1)
    {
        // Interrupt due to PCA Module 1 timeout

        // Reset interrupt source flag
        CCF1 = 0;

        // Mark inter_trans_length invalid
        inter_trans_length = 8;
    }
    else if(CCF2)
    {
        // Interrupt due to PCA Module 2 match

        // We are in one of the transmit states 
        // Toggle transmit output pin
        TX_PORT_PIN ^= 1;

        // Reset interrupt source flag
        CCF2 = 0;
    }

    switch(hdlc_state)
    {        
        case RX_IDLE:
        {
            // Check if any data to transmit
            if (tx_buffer_read_index != tx_buffer_write_index)
// A N D   TX_DWAIT   M A T U R E //
            {
                // Data to transmit

                // Move buffer trial pointer to first packet
                tx_buffer_trial_index = tx_buffer_read_index;

                // Advance buffer trial pointer to data area
                tx_buffer_trial_index++; 
                tx_buffer_trial_index &= TX_BUFFER_LEN;

                // Reset PCA module 2 interrupt source flag
                CCF2 = 0;

                // Enable PCA module 2 interrupt. 
                CCAPM2 |= 0x01;   // Set ECCF2 = 1

                // Disable PCA module 0 & 1.
                // Resetting interrupt enables ECCF0, ECCF1 to 0
                // doesn't seem to be enough to stop them interrupting 
                // during transmission.
                CCAPM0 = 0x00;
                CCAPM1 = 0x00;

                // Reload next interrupt count for 6 bit SYNC
                // Reload PCA module 2 match register
                CCAP2L = compare_reload_L[7];
                CCAP2H = compare_reload_H[7];

                // Move to next state
                hdlc_state = TX_PREAMBLE_0;

                // Turn on transmit output pin
                TX_PORT_PIN = 1;
            }
            // Check for synchronisation sequence of 6 one bits
            else if (inter_trans_length == 7)
            {
                // A sync flag bit sequence of 6 one bits found

                // Dump any data in rx shift register and reset bit counter
                rx_reset_shift_reg();

                hdlc_state = RX_SYNCHRONISED;
            }
        }
        break;

        case RX_SYNCHRONISED:
        {
            // Check for < 1/2 data bit length transition (noise)
            if (inter_trans_length == 0)
            {
                // Wait for next incoming synchronisation sequence
                hdlc_state = RX_IDLE;
                break;
            }
            // Check for valid incoming data
            else if (inter_trans_length > 6) 
            {
                // No valid incoming data detected.
                // Check for any received packets in incoming buffer
                if (rx_buffer_read_index != rx_buffer_write_index)
                {
                    // Unprocessed packets in data buffer so process them now
                    rx_process_packets();
                }
 
                // Wait for next incoming synchronisation sequence
                hdlc_state = RX_IDLE;
                break;
            }
            else // ((inter_trans_length < 7) && (inter_trans_length > 0))
            {
                // inter_trans_length between 1 and 6
                // Incoming valid data detected. Decode it now
                hdlc_state = RX_RECEIVING;

                // No break statement here so we fall straight through into  
                // case RX_RECEIVING: (regardless of new hdlc_state value)
            }
        }

        case RX_RECEIVING:
        {
            // Check for < 1/2 data bit length transition (noise)
            if (inter_trans_length == 0)
            {
                // Incoming data stream lost so discard current packet now
                rx_drop_packet();

                // Wait for incoming synchronisation sequence
                hdlc_state = RX_IDLE;
                break;
            }
            // Check for synchronisation sequence of 6 one bits
            else if (inter_trans_length == 7)
            {
                // Check for any data in current packet buffer
                if (rx_buffer[rx_buffer_write_index] != 0)
                {
                    // Check if sync sequence is byte aligned indicating that
                    // the end of an incoming packet is reached
                    if (rx_bit_count == 1)
                    {
                        // The end of an incoming packet is reached so save 
                        // packet. Move write pointer up to commit packet
                        rx_buffer_write_index = rx_buffer_trial_index;

                        // Zero next incoming packet length
                        rx_buffer[rx_buffer_write_index] = 0;

                        // Advance trial pointer to data area of new packet
                        rx_buffer_trial_index++;

                        // Dump rx shift register data and reset bit counter
                        rx_reset_shift_reg();

                        // Return to synchronised state as this may have been
                        // an interstitial sync between packets
                        hdlc_state = RX_SYNCHRONISED;
                        break;
                    }
                    else
                    {
                        // Sync sequence spurious so discard current packet
                        rx_drop_packet();

                        // Wait for next incoming synchronisation sequence
                        hdlc_state = RX_IDLE;
                        break;
                    }
                }
                else // No recorded data in current packet buffer
                {
                    // Sync sequence is part of initial synch preamble
                    // Dump any data in rx shift register and reset bit counter
                    rx_reset_shift_reg();

                    // Return to synchronised state as this was part of the 
                    // initial synch preamble
                    hdlc_state = RX_SYNCHRONISED;
                    break;
                }
            }
            else if (inter_trans_length < 7)
            {
                // Valid incoming data detected so add it to buffer
                unsigned char ones_count = inter_trans_length - 1;

                // Store incoming ones
                while (ones_count--)
                {
                    rx_add_bit_to_buffer(1);
                }
                
                // Store transition zeros only if they are not the result 
                 // of bit stuffing. ie. Un-stuff incoming stream
                if (inter_trans_length != 6)
                {
                    // A zero detected that didn't occur after 5 ones
                    // Shift a zero into LHS of shift reg
                    rx_add_bit_to_buffer(0);
                }

                // Check for buffer overrun or packet too long 
                if ((rx_buffer_trial_index == rx_buffer_read_index)
                    || (rx_buffer[rx_buffer_write_index] > RX_MAX_PACKET_LEN))
                {
                    // Incoming packet too big or buffer too full
                    // Discard current packet
                    rx_drop_packet();

                    // Wait for incoming synchronisation sequence
                    hdlc_state = RX_IDLE;
                    break;
                }
            }
            else // (inter_trans_length > 7)
            {
                // Incoming data stream lost so discard current packet now
                rx_drop_packet();

                // Wait for incoming synchronisation sequence
                hdlc_state = RX_IDLE;
                break;
            }
        }
        break;

        case TX_PREAMBLE_0:
        {
            // Reload next interrupt count for 1 bit
            // Reload PCA module 2 match register
            CCAP2L = compare_reload_L[1];
            CCAP2H = compare_reload_H[1];

            // Move to next state
            hdlc_state = TX_PREAMBLE_1;
        }
        break;

        case TX_PREAMBLE_1:
        {
            // Check whether any more SYNC to send
            if (--tx_preamble_cnt)
            {
                // At least one more SYNC to send

                // Move to next state
                hdlc_state = TX_PREAMBLE_0;
            }
            else
            {
                // No more SYNC to send

                // Move to next state
                hdlc_state = TX_TRANSMITTING;
            }
            // Reload next interrupt count for 6 bit SYNC
            // Reload PCA module 2 match register
            CCAP2L = compare_reload_L[7];
            CCAP2H = compare_reload_H[7];
        }
        break;

        case TX_TRANSMITTING:
        {
            unsigned char tx_ones_length = 0;  // Length in bit time between 0s
                                               // that enclose a sequence of 
                                               // between none and five 1 bits.
                                               // ie. one to six bit lengths

            // find next zero bit, or five consecutive ones
            do
            {
                // Check for empty tx_shift_reg
                if(!tx_bit_count)
                {        
                    // tx_shift_reg exhausted

                    // Check whether there is any more data in this packet
                    if(tx_buffer[tx_buffer_read_index]-- == 0)
                    {
                        // No more data in this packet

                        // Completed packet, move to next packet start position
                        tx_buffer_read_index = tx_buffer_trial_index;

                        // Advance buffer trial pointer to next data area
                        tx_buffer_trial_index++; 
                        tx_buffer_trial_index &= TX_BUFFER_LEN;

                        // Reload next count for interrupt at first 0 bit of SYNC
                        // Reload PCA module 2 match register
                        CCAP2L = compare_reload_L[tx_ones_length + 1];
                        CCAP2H = compare_reload_H[tx_ones_length + 1];

                        // Move to next state
                        hdlc_state = TX_INTER_SYNC_0;

                        // Re-enable PCA interrupt
                        EC = 1;

                        return;
                    }

                    // Copy next byte to send into tx_shift_reg
                    tx_shift_reg = tx_buffer[tx_buffer_trial_index++];
                    tx_buffer_trial_index &= TX_BUFFER_LEN;
                    tx_bit_count = 7;
                }
                else
                { 
                    // tx_shift_reg not yet exhausted

                    tx_shift_reg >>= 1;
                    tx_bit_count--;
                }
                tx_ones_length++;

            } while((tx_ones_length < 6) && (tx_shift_reg & 1));


            // Five consecutive 1s found so stuff a 0 between 5th and 6th
            if(tx_ones_length == 6)
            {
                tx_bit_count++;
                tx_shift_reg <<= 1;
            }

            // Make next transition
            // Reload next interrupt count
            // Reload PCA module 2 match register
            CCAP2L = compare_reload_L[tx_ones_length];
            CCAP2H = compare_reload_H[tx_ones_length];
        }
        break;

        case TX_INTER_SYNC_0:
        {
            // Check for any more data packets to send
            if (tx_buffer_read_index == tx_buffer_write_index)
            {
                // No more data in tx_buffer to send
                hdlc_state = TX_INTER_SYNC_1;

                // Reset preamble SYNC counter
                tx_preamble_cnt = TX_PREAMBLE_LEN;
            }
            else
            {
                // More data in tx_buffer yet to send
                hdlc_state = TX_TRANSMITTING;
            }

            // Reload next interrupt count for 6 bit SYNC
            // Reload PCA module 2 match register
            CCAP2L = compare_reload_L[7];
            CCAP2H = compare_reload_H[7];
        }
        break;

        case TX_INTER_SYNC_1:
        {
            // Check whether tx pin is high
            if (TX_PORT_PIN)
            {
                // Tx output still high so do another 8 bit long transition

                // Reload next interrupt count for 8 bit 
                // Reload PCA module 2 match register
                CCAP2L = compare_reload_L[8];
                CCAP2H = compare_reload_H[8];

                // No more tx transitions needed
                hdlc_state = TX_INTER_SYNC_2;

                // Re-enable PCA interrupt
                EC = 1;

                return;
            }
            // Tx output pin already low
            // No more tx transitions needed
            hdlc_state = TX_INTER_SYNC_2;
        }
        // No break here so immediately fall through to TX_INTER_SYNC_2
        // regardless of any hdlc_state change

        case TX_INTER_SYNC_2:
        {
            // Wait for more data to send
            hdlc_state = RX_IDLE;

            // Disable PCA module 2 interrupt
            CCAPM2 &= ~0x01;   // Set ECCF2 = 0

            // Reset PCA modules 0 & 1 interrupt source flags
            CCF0 = 0;
            CCF1 = 0;

            // Re-enable PCA module 0 & 1
            CCAPM0 = 0x31;
            CCAPM1 = 0x49;

            // Load PCA module 1 compare registers with count for compare match
            // when an 8 bit long, no transition period occurs.
            CCAP1L = compare_reload_L[8];
            CCAP1H = compare_reload_H[8];
        }
        break;

    } // switch end

    // Re-enable PCA interrupt
    EC = 1;

    return;
}


/*  ͻ
                                                                         
                               rx_process_packets()                      
                                                                         
     Check CRC of each received packet and pass valid packets up stack.  
                                                                         
    ͼ  */
void rx_process_packets(void) 
{
    // For each packet in rx_buffer
    while (rx_buffer_read_index != rx_buffer_write_index)
    {
        // Initialise CRC
        unsigned int CRC = 0xFFFF;

        // Pass each of this packet's data bytes through CRC
        for(rx_buffer_trial_index = rx_buffer_read_index + 1;    
            rx_buffer_trial_index != (rx_buffer_read_index + 
                     rx_buffer[rx_buffer_read_index] - 1) & RX_BUFFER_LEN;
            rx_buffer_trial_index++, rx_buffer_trial_index &= RX_BUFFER_LEN)
        {
            // Pass byte through CRC
            // This isr uses register bank 2 so CRCPass() can't be called here!
            CRC = (CRC >> 8) ^ crc_table[(CRC & 0xFF) 
                             ^ rx_buffer[rx_buffer_trial_index]];
        }

        // Complete CRC computation by inverting all bits
        CRC = ~CRC; 

        // Check if recalculated CRC matches received CRC
        if (CRC == (rx_buffer[rx_buffer_trial_index] 
               + (rx_buffer[(rx_buffer_trial_index + 1) & RX_BUFFER_LEN] << 8)))
        {
            // Packet data passed checksum

            // Do upcall passing address of valid data packet in rx_buffer
            // hdlc_upcall(packet_ptr);
			// Upper layer is responsible for freeing allocated memory

            unsigned char i;
            unsigned char xdata *packet_ptr;

			// Allocate enough memory to hold this packet including length
            packet_ptr = malloc(rx_buffer[rx_buffer_read_index] - 1);

            if(packet_ptr != NULL)
            {
			    // Copy packet length & contents to allocated memory
                for (i = 0; i < rx_buffer[rx_buffer_read_index] - 1; i++)
                {
                    *(packet_ptr + i) = rx_buffer[(rx_buffer_read_index + i)
				                                             & RX_BUFFER_LEN];
                }
                // Remove CRC bytes from length
                *packet_ptr -= 2;

				// Pass packet up to higher layer
                hdlc_upcall(packet_ptr);
            }
        }
        else
        {
            // Packet data failed checksum
            // Increment numeric record of corrupted data
        }

        // Advance the packet buffer trial pointer to next packet start
        rx_buffer_trial_index += 2; 
        rx_buffer_trial_index &= RX_BUFFER_LEN;

        // Move to next packet in rx_buffer
        rx_buffer_read_index = rx_buffer_trial_index;
    }

    // Advance the packet buffer trial pointer to data area
    rx_buffer_trial_index++; 
    rx_buffer_trial_index &= RX_BUFFER_LEN;

    return;
}


/*  ͻ
                                                                         
                                 rx_drop_packet()                        
                                                                         
     Discard any partial packet data in the buffer (if any) and restart  
     from last buffered data packet.                                     
                                                                         
    ͼ  */
void rx_drop_packet(void) 
{
    // Zero incoming packet length
    rx_buffer[rx_buffer_write_index] = 0;

    // Reset incoming packet trial write pointer to data area
    rx_buffer_trial_index = (rx_buffer_write_index + 1) & RX_BUFFER_LEN;

    // Dump any data in rx shift register and reset its bit counter
    rx_reset_shift_reg();
 
    return;
}


/*  ͻ
                                                                         
                            rx_add_bit_to_buffer()                       
                                                                         
     Rotates a zero or a one into the LHS of the rx_shift_reg.           
     If the shift register becomes full, it is written to the rx-buffer. 
                                                                         
    ͼ  */
void rx_add_bit_to_buffer(bit bit_to_add)
{
    // Shift a bit into LHS of shift reg
    rx_shift_reg >>= 1;
    rx_shift_reg |= (unsigned char)bit_to_add << 7;

    // Remember how many bits shifted in
    rx_bit_count++;

    // If shift register full place data in buffer
    if (rx_bit_count == 8)
    {
        // Increment length record for buffered packet
        rx_buffer[rx_buffer_write_index]++;

        // Save the register in buffer
        rx_buffer[rx_buffer_trial_index] = rx_shift_reg;

        // Advance the packet buffer trial pointer
        rx_buffer_trial_index++; 
        rx_buffer_trial_index &= RX_BUFFER_LEN;

        // Dump any data in rx shift register and reset its bit counter
        rx_reset_shift_reg();
    }

    return;
}


/*  ͻ
                                                                         
                              rx_reset_shift_reg()                       
                                                                         
     Clear any data in rx shift register and reset it's bit counter.     
                                                                         
    ͼ  */
void rx_reset_shift_reg(void)
{
    // Clear rx shift register
    rx_shift_reg = 0;

    // Reset rx shift register bit counter
    rx_bit_count = 0;

    return;
}


/*  ͻ
                                                                         
                                 tx_send_packet()                        
                                                                         
                                                                         
                                                                         
    ͼ  */
bit tx_send_packet(unsigned char * packet)
{
    bit error = 0;

    // Initialise CRC
    unsigned int CRC = 0xFFFF;

    // Reset length record for buffered packet
    tx_buffer[tx_buffer_write_index] = 0;       

    // Pass each of this packet's data bytes through CRC
    while ((*packet) && (!(error = tx_buffer_overrun())))
    {
        // Pass byte through CRC
        CRC = CRCPass(CRC, *packet);

        // Add byte to tx_buffer and move to next packet byte
        tx_buffer[tx_buffer_trial_index] = *packet++;

        // Advance the packet buffer trial pointer to next byte
        tx_buffer_trial_index++; 
        tx_buffer_trial_index &= TX_BUFFER_LEN;

        // Increment length record for buffered packet
        tx_buffer[tx_buffer_write_index]++;       
    }

    if (!error)
    {
        // Packet data successfully entered into buffer

        // Complete CRC computation by inverting all bits
        CRC = ~CRC; 

        // Add CRC low byte to tx_buffer
        tx_buffer[tx_buffer_trial_index] = (CRC & 0xFF);

        // Advance the packet buffer trial pointer to next packet start
        tx_buffer_trial_index++; 
        tx_buffer_trial_index &= TX_BUFFER_LEN;

        // Increment length record for buffered packet
        tx_buffer[tx_buffer_write_index]++;       

        // Check for buffer too full
        error = tx_buffer_overrun();
    }

    if (!error)
    {
        // Add CRC high byte to tx_buffer
        tx_buffer[tx_buffer_trial_index] = (CRC >> 8);

        // Advance the packet buffer trial pointer to next packet start
        tx_buffer_trial_index++; 
        tx_buffer_trial_index &= TX_BUFFER_LEN;

        // Increment length record for buffered packet
        tx_buffer[tx_buffer_write_index]++;       
    }

    if (!(error = tx_buffer_overrun()))
    {
        // Commit packet and move to next packet start position
        tx_buffer_write_index = tx_buffer_trial_index;

        // Advance the packet buffer trial pointer to data area
        tx_buffer_trial_index++; 
        tx_buffer_trial_index &= TX_BUFFER_LEN;
    }

    if (error)
    {
        // An error occured so reset current buffer

        // Restart trial pointer at beginning of packet data area
        tx_buffer_trial_index = tx_buffer_write_index + 1;

        // Reset current packet length record;
        tx_buffer[tx_buffer_write_index] = 0;
    }

    return error;
}



/*  ͻ
                                                                         
                               tx_buffer_overrun()                       
                                                                         
                                                                         
                                                                         
    ͼ  */
bit tx_buffer_overrun(void)
{
    return ((tx_buffer_trial_index == tx_buffer_read_index)
               || (tx_buffer[tx_buffer_write_index] > TX_MAX_PACKET_LEN));
}
